r"""
GdSymbolTable class
"""
from enum import Enum, auto

from gdoc.lib.gdoccompiler.gdexception import GdocIdError, GdocRuntimeError, GdocTypeError

from .gdsymbol import GdSymbol


class GdSymbolTable:
    """
    ;
    """

    class Type(Enum):
        """Symboltable element classes"""

        OBJECT = auto()
        REFERENCE = auto()
        IMPORT = auto()
        ACCESS = auto()

    def __init__(self, id, *, scope="+", name=None, tags=[], _type=Type.OBJECT):
        """split symbol string to ids or names and tags.
        @param symbol : str | PandocStr
        @return (list(str), list(str))
        """
        if (id is not None) and (id == ""):
            raise GdocIdError('invalid id ""')

        elif (id is not None) and id.startswith(GdSymbol.IS_NAME_STR):
            raise GdocIdError('invalid id "' + id + '"')

        elif scope not in ("+", "-"):
            raise GdocRuntimeError('invalid access modifier "' + scope + '"')

        elif type(tags) is not list:
            raise TypeError("only a list can be added as tags")

        elif type(_type) is not GdSymbolTable.Type:
            raise TypeError("only GdSymbolTable.Type can be set")

        elif (name is not None) and (name == ""):
            raise GdocIdError('invalid name ""')

        if (id is None) and (name is None):
            raise GdocIdError("at least one of the id or name is required")

        if _type is GdSymbolTable.Type.IMPORT:
            scope = "+"
        elif _type is GdSymbolTable.Type.ACCESS:
            scope = "-"

        self.id = id
        self.scope = scope
        self.name = name
        self.tags = tags[:]

        self.__type = _type
        self.__parent = None
        self.__children = []
        self.__references = []
        self.__idlist = {}
        self.__namelist = {}
        self.__cache = []
        self.__link_to = None
        self.__link_from = []

    def add_child(self, child: "GdSymbolTable"):
        """split symbol string to ids or names and tags.
        @param symbol : str | PandocStr
        @return (list(str), list(str))
        """
        if self.__type is GdSymbolTable.Type.IMPORT:
            raise GdocTypeError("'Import' type cannot have children")

        elif self.__type is GdSymbolTable.Type.ACCESS:
            raise GdocTypeError("'Access' type cannot have children")

        elif child.__type is GdSymbolTable.Type.REFERENCE:
            self.__references.append(child)
            child.__parent = self

        else:
            child.__parent = self

            if child.id is not None:
                if str(child.id) in self.__idlist:
                    raise GdocIdError('duplicate id "' + child.id + '"')
                self.__idlist[str(child.id)] = child

            if child.name is not None:
                if str(child.name) in self.__namelist:
                    raise GdocIdError('duplicate name "' + child.name + '"')
                self.__namelist[str(child.name)] = child

            self.__children.append(child)

    def get_parent(self):
        """split symbol string to ids or names and tags.
        @param symbol : str | PandocStr
        @return (list(str), list(str))
        """
        return self.__parent

    def get_children(self):
        """split symbol string to ids or names and tags.
        @param symbol : str | PandocStr
        @return (list(str), list(str))
        """
        children = []

        parents = self.__get_linkto_target().__get_linkfrom_list()

        for parent in parents:
            children += parent.__get_children()

        return children

    def get_child(self, id):
        """split symbol string to ids or names and tags.
        @param symbol : str | PandocStr
        @return (list(str), list(str))
        """
        child = None

        parents = self.__get_linkto_target().__get_linkfrom_list()

        for parent in parents:
            if id in parent.__idlist:
                child = parent.__idlist[id]
                break

        return child

    def get_child_by_name(self, name):
        """split symbol string to ids or names and tags.
        @param symbol : str | PandocStr
        @return (list(str), list(str))
        """
        child = None

        parents = self.__get_linkto_target().__get_linkfrom_list()

        for parent in parents:
            if name in parent.__namelist:
                child = parent.__namelist[name]
                break

        return child

    def __get_children(self):
        """split symbol string to ids or names and tags.
        @param symbol : str | PandocStr
        @return (list(str), list(str))
        """
        return self.__children[:]

    def __get_references(self):
        """split symbol string to ids or names and tags.
        @param symbol : str | PandocStr
        @return (list(str), list(str))
        """
        return self.__references[:]

    def resolve(self, symbols):
        """split symbol string to ids or names and tags.
        @param symbol : str | PandocStr
        @return (list(str), list(str))
        """
        target = None

        if symbols[0].startswith("*"):
            target = self.get_child_by_name(symbols[0][1:])
        else:
            target = self.get_child(symbols[0])

        if target is not None:
            return target

        root = self
        while root is not None:
            if symbols[0].startswith("*"):
                if root.name == symbols[0][1:]:
                    target = root
                    break
            else:
                if root.id == symbols[0]:
                    target = root
                    break

            root = root.get_parent()

        if target is None:
            return 0

        for i in range(1, len(symbols)):
            if symbols[i].startswith("*"):
                target = root.get_child_by_name(symbols[i][1:])
            else:
                target = root.get_child(symbols[i])

            if target is None:
                target = i
                break

            root = target

        return target

    def unidir_link_to(self, dst: "GdSymbolTable"):
        """split symbol string to ids or names and tags.
        @param symbol : str | PandocStr
        @return (list(str), list(str))
        """
        if self.__type is GdSymbolTable.Type.OBJECT:
            raise TypeError("'OBJECT' cannot unidir_link to any others")

        elif self.__type is GdSymbolTable.Type.REFERENCE:
            raise TypeError("'REFERENCE' cannot unidir_link to any others")

        self.__link_to = dst

    def bidir_link_to(self, dst: "GdSymbolTable"):
        """split symbol string to ids or names and tags.
        @param symbol : str | PandocStr
        @return (list(str), list(str))
        """
        if self.__type is GdSymbolTable.Type.REFERENCE:

            if dst.__type in (GdSymbolTable.Type.OBJECT, GdSymbolTable.Type.REFERENCE):
                self.__link_to = dst
                dst.__link_from.append(self)

            elif dst.__type is GdSymbolTable.Type.IMPORT:
                raise TypeError("cannot bidir_link to 'IMPORT'")

            else:  # dst.__type is GdSymbolTable.Type.ACCESS:
                raise TypeError("cannot bidir_link to 'ACCESS'")

        elif self.__type is GdSymbolTable.Type.OBJECT:
            raise TypeError("'OBJECT' cannot bidir_link to any others")

        elif self.__type is GdSymbolTable.Type.IMPORT:
            raise TypeError("'IMPORT' cannot bidir_link to any others")

        else:  # self.__type is GdSymbolTable.Type.ACCESS:
            raise TypeError("'ACCESS' cannot bidir_link to any others")

    def __get_linkto_target(self):
        """split symbol string to ids or names and tags.
        @param symbol : str | PandocStr
        @return (list(str), list(str))
        """
        target = self

        while target.__type is not GdSymbolTable.Type.OBJECT:
            target = target.__link_to
            if target is None:
                break

        return target

    def __get_linkfrom_list(self):
        """split symbol string to ids or names and tags.
        @param symbol : str | PandocStr
        @return (list(str), list(str))
        """
        _list = [self]

        for item in self.__link_from:
            _list += item.__get_linkfrom_list()

        return _list



# SysML: System Description Language plugin
#
# 1. Block definition diagram の、参照関係をのぞくブロック単体定義相当の情報を取り扱う
#

import logging

from ...lib import debug
from ...lib.types.table.hierarchicalDict import HierarchicalDict

_LOGGER = logging.getLogger(__name__)
_DEBUG = debug.Debug(_LOGGER)

_version = "0.2.0"


class Block(HierarchicalDict):
    def __init__(self, table, tag=None, parent=None) -> None:
        _DEBUG.print("class Block(HierarchicalDict) {")
        _DEBUG.indent()

        super().__init__(table, tag)

        self.content[0]["plugin"] = "sysml"
        self.content[0]["type"] = "block"
        self.content[0]["version"] = _version

        _DEBUG.undent()
        _DEBUG.print("}")



"""
command.py
"""
import json

from gdoc.lib.gdoc import String, Text
from gdoc.lib.gdoccompiler.gdcompiler.gdcompiler import GdocCompiler
from gdoc.util import ErrorReport


def setup(subparsers, name, commonOptions):
    """
    Setup subcommand
    """
    global __subcommand__
    __subcommand__ = name

    parser = subparsers.add_parser(
        __subcommand__,
        parents=[commonOptions],
        help="Compile a markdown file to a gdoc object file.",
    )
    parser.set_defaults(func=run)
    parser.add_argument("filepath", help="target source file", nargs="+")
    parser.add_argument(
        "-f", "--filetype", help="Input filetype(to be specified for pandoc)"
    )
    parser.add_argument(
        "--html", action="store_true", help="Interprete HTML tags when parsing markdown."
    )
    parser.add_argument(
        "--check-only",
        action="store_true",
        help="Performs only syntax checking on the document.",
    )


def run(args):
    """
    run subcommand
    """
    erpt: ErrorReport

    for filepath in args.filepath:
        gobj, erpt = GdocCompiler().compile(
            filepath, opts={"check-only": args.check_only}
        )

        if gobj is not None:
            data = _export_gobj(gobj)
            print(json.dumps(data, indent=4, ensure_ascii=False))

        if erpt is not None:
            print(erpt.dump())


def _export_gobj(gobj):
    prop = gobj._GdObject__properties
    data = _cast_to_str(prop)

    data["children"] = []
    children = gobj.get_children()
    for child in children:
        data["children"].append(_export_gobj(child))

    return data


def _cast_to_str(prop):

    if isinstance(prop, dict):
        keys = prop.keys()
    elif isinstance(prop, list):
        keys = range(len(prop))

    for key in keys:
        if isinstance(prop[key], String):
            prop[key] = str(prop[key])

        elif isinstance(prop[key], Text):
            prop[key] = str(prop[key].get_str())

        elif isinstance(prop[key], (dict, list)):
            prop[key] = _cast_to_str(prop[key])

    return prop



r"""
PandocAst Type definitions and Utilities.
"""

from .blocklist import BlockList
from .element import Element
from .inline import Inline
from .inlinelist import InlineList
from .pandoc import Pandoc


def create_element(pan_elem, elem_type=None) -> Element:
    """
    Find the element type and call constructor specified by it.
    """
    etype = "ELEMENT TYPE MISSING"

    if elem_type is not None:
        etype = elem_type

    elif "t" in pan_elem:
        etype = pan_elem["t"]

    elif "pandoc-api-version" in pan_elem:
        etype = "Pandoc"

    if etype not in _ELEMENT_TYPES:
        # Invalid etype( = 'ELEMENT TYPE MISSING' or invalid `elem_type`)
        raise KeyError(etype)

    element = _ELEMENT_TYPES[etype]["class"](
        pan_elem, etype, _ELEMENT_TYPES[etype], create_element
    )

    return element


class PandocAst(Pandoc):
    def __init__(self, pan_elem):
        """Constructor
        @param pan_elem(Dict)
            PandocAST Element
        """
        super().__init__(pan_elem, "Pandoc", _ELEMENT_TYPES["Pandoc"], create_element)

    create_element = create_element


# Text.Pandoc.Definition
# Definition of Pandoc data structure for format-neutral representation of documents.
# https://hackage.haskell.org/package/pandoc-types-1.22/docs/Text-Pandoc-Definition.html
#
_ELEMENT_TYPES = {
    #
    # Gdoc additional types
    #
    "Line": {
        # [Inline]  is not Line object, just an Array of Inlines.
        # It's 'Multiple non-breaking lines' for LineBlock class.
        "class": InlineList,
        "content": {"key": None, "type": None},
        "struct": None,
        "separator": "",
    },
    "ListItem": {
        # [Block]   is not ListItem object, just an Array of Blocks.
        "class": BlockList,
        "content": {"key": None, "type": None},
        "struct": None,
    },
    #
    # Pandoc
    #
    "Pandoc": {
        # Pandoc Meta [Block]
        "class": Pandoc,
        "content": {"key": None, "main": "blocks", "type": None},
        "struct": {"Version": "pandoc-api-version", "Meta": "meta", "Blocks": "blocks"},
    },
    #
    # Blocks
    #
    "Plain": {
        # Plain [Inline]
        # - Plain text, not a paragraph
        "class": InlineList,
        "content": {"key": "c", "type": None},
        "struct": None,
        "separator": "",
    },
    "Para": {
        # Para [Inline]
        # - Paragraph
        "class": InlineList,
        "content": {"key": "c", "type": None},
        "struct": None,
        "separator": "",
    },
    "LineBlock": {
        # LineBlock [[Inline]]
        # - Multiple non-breaking lines
        "class": InlineList,
        "content": {"key": "c", "type": "Line"},
        "struct": None,
        "separator": "\n",
    },
    "CodeBlock": {
        # CodeBlock Attr Text
        # - Code block (literal) with attributes
        "class": InlineList,
        "content": {"key": "c", "main": 1, "type": "Text"},
        "struct": {"Attr": 0, "Text": 1},
    },
    "RawBlock": {
        # RawBlock Format Text
        # - Raw block
        "class": InlineList,
        "content": {"key": "c", "main": 1, "type": "Text"},
        "struct": {"Format": 0, "Text": 1},
    },
    "BlockQuote": {
        # BlockQuote [Block]
        # - Block quote (list of blocks)
        "class": BlockList,
        "content": {"key": "c", "type": None},
        "struct": None,
    },
    "OrderedList": {
        # OrderedList ListAttributes [[Block]]
        # - Ordered list (attributes and a list of items, each a list of blocks)
        "class": BlockList,
        "content": {"key": "c", "main": 1, "type": "ListItem"},
        "struct": {"ListAttributes": 0, "ListItems": 1},
    },
    "BulletList": {
        # BulletList [[Block]]
        # - Bullet list (list of items, each a list of blocks)
        "class": BlockList,
        "content": {"key": "c", "type": "ListItem"},
        "struct": None,
    },
    "Header": {
        # Header Int Attr [Inline]
        # - Header - level (integer) and text (inlines)
        "class": InlineList,
        "content": {"key": "c", "main": 2, "type": None},
        "struct": {"Level": 0, "Attr": 1, "[Inline]": 2},
        "separator": "",
    },
    "HorizontalRule": {
        # HorizontalRule
        # - Horizontal rule
        "class": InlineList,
        "alt": "",
    },
    # 'Table':  {
    #     # Table Attr Caption [ColSpec] TableHead [TableBody] TableFoot
    #     # - Table, with attributes, caption, optional short caption, column alignments and widths (required), table head, table bodies, and table foot
    #     'class':  Table,
    #     'content':  {
    #         'key':      'c',
    #         'main':     None
    #     },
    #     'struct': {
    #         'Attr':         0,
    #         'Caption':      1,
    #         '[ColSpec]':    2,
    #         'TableHead':    3,
    #         '[TableBody]':  4,      # TableBody*s*
    #         'TableFoot':    5
    #     }
    # },
    "Div": {
        # Div Attr [Block]
        # - Generic block container with attributes
        "class": BlockList,
        "content": {"key": "c", "main": 1, "type": None},
        "struct": {"Attr": 0, "[Block]": 1},
    },
    #
    # Inlines
    #
    "Str": {
        # Str Text
        # Text (string)
        "class": Inline,
        "content": {"key": "c", "type": "Text"},
        "struct": None,
    },
    "Emph": {
        # Emph [Inline]
        # Emphasized text (list of inlines)
        "class": Inline,
        "content": {"key": "c", "type": None},
        "struct": None,
    },
    "Underline": {
        # Underline [Inline]
        # Underlined text (list of inlines)
        "class": Inline,
        "content": {"key": "c", "type": None},
        "struct": None,
    },
    "Strong": {
        # Strong [Inline]
        # Strongly emphasized text (list of inlines)
        "class": Inline,
        "content": {"key": "c", "type": None},
        "struct": None,
    },
    "Strikeout": {
        # Strikeout [Inline]
        # Strikeout text (list of inlines)
        "class": Inline,
        "content": {"key": "c", "type": None},
        "struct": None,
    },
    "Superscript": {
        # Superscript [Inline]
        # Superscripted text (list of inlines)
        "class": Inline,
        "content": {"key": "c", "type": None},
        "struct": None,
    },
    "Subscript": {
        # Subscript [Inline]
        # Subscripted text (list of inlines)
        "class": Inline,
        "content": {"key": "c", "type": None},
        "struct": None,
    },
    "SmallCaps": {
        # SmallCaps [Inline]
        # Small caps text (list of inlines)
        "class": Inline,
        "content": {"key": "c", "type": None},
        "struct": None,
    },
    "Quoted": {
        # Quoted QuoteType [Inline]
        # Quoted text (list of inlines)
        "class": Inline,
        "content": {"key": "c", "main": 1, "type": None},
        "struct": {"QuotedType": 0, "Inlines": 1},
    },
    "Cite": {
        # Cite [Citation] [Inline]
        # Citation (list of inlines)
        "class": Inline,
        "content": {"key": "c", "main": 1, "type": None},
        "struct": {"Citation": 0, "Inlines": 1},
    },
    "Code": {
        # Code Attr Text
        # Inline code (literal)
        "class": Inline,
        "content": {"key": "c", "main": 1, "type": "Text"},
        "struct": {"Attr": 0, "Text": 1},
    },
    "Space": {
        # Space
        # Inter-word space
        "class": Inline,
        "alt": " ",
    },
    "SoftBreak": {
        # SoftBreak
        # Soft line break
        "class": Inline,
        "alt": " ",
    },
    "LineBreak": {
        # LineBreak
        # Hard line break
        "class": Inline,
        "alt": "\n",
    },
    "Math": {
        # Math MathType Text
        # TeX math (literal)
        "class": Inline,
        "content": {"key": "c", "main": 1, "type": "Text"},
        "struct": {"MathType": 0, "Text": 1},
    },
    "RawInline": {
        # RawInline Format Text
        # Raw inline
        "class": Inline,
        "content": {"key": "c", "main": 1, "type": "Text"},
        "struct": {"Format": 0, "Text": 1},
    },
    "Link": {
        # Link Attr [Inline] Target
        # Hyperlink: alt text (list of inlines), target
        "class": Inline,
        "content": {"key": "c", "main": 1, "type": None},
        "struct": {"Attr": 0, "Inlines": 1, "Target": 2},
    },
    "Image": {
        # Image Attr [Inline] Target
        # Image: alt text (list of inlines), target
        "class": Inline,
        "content": {"key": "c", "main": 1, "type": None},
        "struct": {"Attr": 0, "Inlines": 1, "Target": 2},
    },
    "Note": {
        # Note [Block]
        # Footnote or endnote
        "class": Inline,
        "content": {"key": "c", "type": None},
        "struct": None,
    },
    "Span": {
        # Span Attr [Inline]
        # Generic inline container with attributes
        "class": Inline,
        "content": {"key": "c", "main": 1, "type": None},
        "struct": {"Attr": 0, "Inlines": 1},
    },
    # #
    # # OtherTypes
    # #
    # 'TableHead':  {
    #     # TableHead Attr [Row]
    #     # The head of a table.
    #     'class':  TableRowList,
    #     'content':  {
    #         'key':      None,
    #         'main':     1,
    #         'type':     '[Row]'
    #     },
    #     'struct': {
    #         'Attr':     0,
    #         'Rows':     1
    #     }
    # },
    # 'TableBody':  {
    #     # TableBody Attr RowHeadColumns [Row] [Row]
    #     # A body of a table, with an intermediate head, intermediate body,
    #     # and the specified number of row header columns in the intermediate body.
    #     'class':  TableBody,
    #     'content':  {
    #         'key':      None
    #     },
    #     'struct': {
    #         'Attr':             0,
    #         'RowHeadColumns':   1,
    #         'RowHeads': {
    #             'index':       2,
    #             'type':         '[Row]'
    #         },
    #         'Rows': {
    #             'index':       3,
    #             'type':         '[Row]'
    #         },
    #     }
    # },
    # 'TableFoot':  {
    #     # TableFoot Attr [Row]
    #     # The foot of a table.
    #     'class':  TableRowList,
    #     'content':  {
    #         'key':      None,
    #         'main':     1,
    #         'type':     '[Row]'
    #     },
    #     'struct': {
    #         'Attr':     0,
    #         'Rows':     1
    #     }
    # },
    # 'Rows':  {
    #     # Row Attr [Cell]
    #     # A table row.
    #     'class':  TableRowList,
    #     'content':  {
    #         'key':      None,
    #         'type':     '[Row]'
    #     },
    #     'struct':       None
    # },
    # 'Row':  {
    #     # Row Attr [Cell]
    #     # A table row.
    #     'class':  TableRow,
    #     'content':  {
    #         'key':      None,
    #         'main':     1,
    #         'type':     '[Cell]'
    #     },
    #     'struct': {
    #         # 'Attr':     0,    Commented out because of issue about handling Rows in Table.
    #         'Cells':    1
    #     }
    # },
    # 'Cell':  {
    #     # Cell Attr Alignment RowSpan ColSpan [Block]
    #     # A table cell.
    #     'class':  TableCell,
    #     'content':  {
    #         'key':      None,
    #         'main':     4,
    #         'type':     '[Block]'
    #     },
    #     'struct': {
    #         'Attr':         0,
    #         'Alignment':    1,
    #         'RowSpan':      2,
    #         'ColSpan':      3,
    #         '[Block]':  {
    #             'index':   4,
    #             'type':     '[Block]'
    #         }
    #     }
    # }
}



"""
command.py
"""
import json
import logging
import sys

from gdoc.lib.pandocast import pandocast

from ...lib import debug, gdom, plugin
from ...lib.pandoc import Pandoc

_LOGGER = logging.getLogger(__name__)
_DEBUG = debug.Debug(_LOGGER)


def setup(subparsers, name, commonOptions):
    """
    Setup subcommand
    """
    global __subcommand__
    __subcommand__ = name

    parser = subparsers.add_parser(
        __subcommand__, parents=[commonOptions], help="show trace tree"
    )
    parser.set_defaults(func=run)
    parser.add_argument("id", help="target id to trace")
    parser.add_argument("filePath", help="display a square of a given number", nargs="*")
    parser.add_argument(
        "--upper", help="enable print debug information.", type=int, default=1
    )
    parser.add_argument(
        "--lower", help="enable print debug information.", type=int, default=1
    )
    parser.add_argument(
        "--long", action="store_true", help="enable print debug information."
    )
    parser.add_argument(
        "--verbose", action="store_true", help="enable print debug information."
    )


def run(args):
    """
    run subcommand
    """
    #
    #  Get pandoc AST json file
    #
    if len(args.filePath) > 0:

        if args.filePath[0].endswith(".json"):
            with open(args.filePath[0], "r", encoding="UTF-8") as f:
                pandoc = json.load(f)

        else:
            p = Pandoc()
            panAST = p.run(args.filePath[0])
            pandoc = json.loads(panAST)

    elif not sys.stdin.isatty():
        # file名の指定がない場合は、パイプから読み込みます。
        # 標準入力が tty ではないことを確認して、標準入力をパイプと判定しています。
        pandoc = json.load(sys.stdin)

    else:
        print(
            __subcommand__
            + ": error: Missing pandocfile ( [-d / --pandocfile] is required)"
        )
        sys.exit(1)

    gdoc = pandocast.PandocAst(pandoc)

    types = plugin.Plugins()
    ghost = gdom.GdocObjectModel(gdoc, types)
    ghost.walk(_link)

    filename = args.filePath[0] if len(args.filePath) > 0 else ""
    items = ghost.symbolTable.search(args.id)
    for item in items:
        _treeView(item, args.upper, args.lower, filename)


def _link(element):
    if hasattr(element, "link") and callable(element.link):
        element.link()


def _treeView(item, upper, lower, filename):
    lines = []
    indent = 0

    # LinkTo(Upper)
    branches = _retrieve(item, "to", upper)
    lines = _format(branches, "to", "")

    # TargetItem
    lines.append(_getItemInfo(item))

    # LinkFrom(Lower)
    branches = _retrieve(item, "from", lower)
    lines += _format(branches, "from", "")

    for line in lines:
        lineString = filename + ":  " + line
        print(lineString)


def _retrieve(item, dir, limit):
    branches = {}

    if not isinstance(item, str):
        for childType in item.link[dir]:
            branches[childType] = []

            for linkItem in item.link[dir][childType]:
                branches[childType].append(
                    {"info": _getItemInfo(linkItem), "children": {}}
                )
                if (limit > 1) or (limit < 0):
                    branches[childType][-1]["children"] = _retrieve(
                        linkItem, dir, limit - 1
                    )

    return branches


def _format(branches, dir, leadString):
    lines = []
    indent = 0
    indentString = ""
    ANGLE = "┌" if dir == "to" else "└"

    c = len(branches)
    for childTyp in branches:
        c -= 1
        if c > 0:
            line = leadString + "├  "
            nextString = leadString + "│   "
        else:
            line = leadString + ANGLE + "  "
            nextString = leadString + "    "

        line += "@" + childTyp
        lines.append(line)

        i = len(branches[childTyp])
        for item in branches[childTyp]:
            i -= 1
            if i > 0:
                line = nextString + "├  "
                childString = nextString + "│  "
            else:
                line = nextString + ANGLE + "  "
                childString = nextString + "   "

            line += item["info"]
            lines.append(line)
            lines += _format(item["children"], dir, childString)

    if dir == "to":
        lines.reverse()

    return lines


def _getItemInfo(item):
    lineString = "SysML.Reqt "

    if isinstance(item, str):
        lineString += item + " - [NOT FOUND]"

    else:
        lineString += item.symboltable.fullId + "." + item.id

        if hasattr(item, "tags") and (len(item.tags) > 0):
            lineString += "(" + ",".join(item.tags) + ")"

        if item.name is not None:
            lineString += " - " + item.name

        # lineString +=  ' : ' + item.text

    return lineString



"""
sectionparser.py: parse_Section function
"""
from gdoc.lib.gdoc import Section, TextBlock
from gdoc.lib.gobj.types import BaseObject
from gdoc.util import Err, Ok, Result

from ...util.errorreport import ErrorReport
from .textblock.textblockparser import parse_TextBlock


def parse_Section(
    section: Section, gobj: BaseObject, opts: dict, erpt: ErrorReport
) -> Result[BaseObject, ErrorReport]:
    """ """
    srpt: ErrorReport = erpt.new_subreport()
    context: BaseObject = gobj
    num_blocks: int = len(section)
    i = 0

    if num_blocks > 0:
        if type(section[0]) is TextBlock:
            #
            # The first block
            #
            context, e = parse_TextBlock(section[0], context, opts, srpt)
            if e and srpt.submit(e):
                return Err(srpt)

            context = context or gobj
            i += 1

        while i < num_blocks:
            #
            # Following blocks
            #
            blocktype = type(section[i])
            if blocktype is TextBlock:
                _, e = parse_TextBlock(section[i], context, opts, srpt)
                if e and srpt.submit(e):
                    return Err(srpt)

            elif blocktype is Section:
                _, e = parse_Section(section[i], context, opts, srpt)
                if e and srpt.submit(e):
                    return Err(srpt)

            i += 1

    if srpt.haserror():
        return Err(srpt, gobj)

    return Ok(gobj)



r"""
ImportObject class
"""

from gdoc.lib.gdoccompiler.gdexception import *

from ..gdsymboltable import GdSymbolTable
from .baseobject import BaseObject


class ImportObject(BaseObject):
    """ """

    def __init__(
        self, typename, id, *, scope=None, name=None, tags=[], ref=None, type_args={}
    ):
        if typename == "IMPORT":
            typename = GdSymbolTable.Type.IMPORT
            if scope not in ("+", None):
                raise GdocRuntimeError()
            scope = "+"

        elif typename == "ACCESS":
            typename = GdSymbolTable.Type.ACCESS
            if scope not in ("-", None):
                raise GdocRuntimeError()
            scope = "-"

        else:
            raise GdocRuntimeError()

        if ref is not None:
            raise GdocRuntimeError()

        super().__init__(
            typename, id, scope=scope, name=name, tags=tags, ref=ref, type_args=type_args
        )



"""
command.py
"""
import json
import logging
import sys

from gdoc.lib.pandocast import pandocast

from ...lib import debug, gdom, plugin
from ...lib.pandoc import Pandoc

_LOGGER = logging.getLogger(__name__)
_DEBUG = debug.Debug(_LOGGER)


def setup(subparsers, name, commonOptions):
    """
    Setup subcommand
    """
    global __subcommand__
    __subcommand__ = name

    parser = subparsers.add_parser(
        __subcommand__, parents=[commonOptions], help="dump Gdoc object"
    )
    parser.set_defaults(func=run)
    parser.add_argument("-p", "--pandocfile", help="path to pandoc AST file.")
    parser.add_argument("-i", "--id", help="id to find and dump.")


def run(args):
    """
    run subcommand
    """
    #
    # dumpdata を取得
    #
    if args.pandocfile is not None:
        if args.pandocfile.endswith(".json"):
            with open(args.pandocfile, "r", encoding="UTF-8") as f:
                pandoc = json.load(f)

        else:
            p = Pandoc()
            panAST = p.run(args.pandocfile)
            pandoc = json.loads(panAST)

    elif not sys.stdin.isatty():
        # file名の指定がない場合は、パイプから読み込みます。
        # 標準入力が tty ではないことを確認して、標準入力をパイプと判定しています。
        # $ pandoc -f gfm+sourcepos -t html ../ghost/process/README.md | pandoc -f html -t json | gdoc pandoc
        pandoc = json.load(sys.stdin)

    else:
        print(
            __subcommand__
            + ": error: Missing pandocfile ( [-d / --pandocfile] is required)"
        )
        sys.exit(1)

    gdoc = pandocast.PandocAst(pandoc)
    gdoc.walk(_dump_gdoc, post_action=_dump_post_gdoc)

    types = plugin.Plugins()
    ghost = gdom.GdocObjectModel(gdoc, types)

    if args.id is None:
        data = {}
        data["Gdoc object"] = ghost.dump()
        data["Namespaces and Items"] = ghost.symbolTable.dump()
        print(json.dumps(data, indent=4, ensure_ascii=False))

    else:
        items = ghost.symbolTable.search(args.id)
        for item in items:
            data = item.getItem()
            del data["objectClass"]
            del data["item"]
            del data["scope"]
            print(json.dumps(data, indent=4, ensure_ascii=False))


def _dump_gdoc(elem, gdoc):
    pos = elem.source.position if elem.source.position is not None else "None"
    if elem.type == "Cell":
        pos = pos + " C" + str(elem.colSpan)
        pos = pos + " R" + str(elem.rowSpan)
    pos = " (" + pos + ")"
    _DEBUG.print(elem.type + pos + " {")
    pass


def _dump_post_gdoc(elem, gdoc):
    if hasattr(elem, "text") and (elem.text is not None):
        if isinstance(elem.text, list):
            _DEBUG.print(">> " + ("\n" + ">> ").join(elem.text), 1)
        else:
            # _DEBUG.print(elem.type + ': ' + elem.text)
            pass

    _DEBUG.print("}")



"""
section.py: Section class
"""

import copy

from ..pandocastobject.pandocast.element import Element
from .config import DEFAULTS
from .textstring import TextString


class TextBlock(list):
    """ """

    def __init__(self, textblock=None, opts={}):
        super().__init__()
        self.__element = textblock
        # self.__opts = opts  # not yet copy() / copy.deepcopy()
        self.__opts = copy.deepcopy(DEFAULTS)
        self.__opts.update(opts)

        ignore: list = self.__opts.get("pandocast", {}).get("types", {}).get("ignore", [])
        ignore += self.__opts.get("pandocast", {}).get("types", {}).get("decorator", [])
        inlines = textblock.get_child_items(ignore=ignore)

        remove: list = self.__opts.get("pandocast", {}).get("types", {}).get("remove", [])
        item: Element = None
        line: list = []
        for item in inlines:
            type = item.get_type()
            if type in remove:
                pass
            elif type == "LineBreak":
                line.append(item)
                self.append(TextString(line, opts=self.__opts))
                line = []
            else:
                line.append(item)

        if len(line) > 0:
            self.append(TextString(line, opts=self.__opts))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gdoc',
 'gdoc.app',
 'gdoc.app.compile',
 'gdoc.app.dump',
 'gdoc.app.trace',
 'gdoc.lib',
 'gdoc.lib.gdoc',
 'gdoc.lib.gdoccompiler',
 'gdoc.lib.gdoccompiler.gdcompiler',
 'gdoc.lib.gdocparser',
 'gdoc.lib.gdocparser.tag',
 'gdoc.lib.gdocparser.textblock',
 'gdoc.lib.gobj',
 'gdoc.lib.gobj.types',
 'gdoc.lib.pandocast',
 'gdoc.lib.pandocastobject',
 'gdoc.lib.pandocastobject.pandoc',
 'gdoc.lib.pandocastobject.pandocast',
 'gdoc.lib.pandocastobject.pandocstr',
 'gdoc.lib.types',
 'gdoc.lib.types.table',
 'gdoc.plugin',
 'gdoc.plugin.sysml',
 'gdoc.util']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['gdoc = gdoc.__main__:main']}

setup_kwargs = {
    'name': 'gdoc',
    'version': '0.1.1',
    'description': 'A tool to process GDML(GDoc Markup Language) documents.',
    'long_description': 'None',
    'author': 'Tsuyoshi Kodama',
    'author_email': 'tsuyoshi.kodama@byrnison.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"""
result.py: Result related types
"""
from typing import Final, Generic, Optional, TypeAlias, TypeGuard, TypeVar

T = TypeVar("T")
E = TypeVar("E")


class Ok(tuple, Generic[T]):
    is_ok: Final[TypeGuard["Ok"]] = True
    is_err: Final[TypeGuard["Err"]] = False

    def __new__(cls, t: T):
        return tuple.__new__(cls, (t, None))

    def ok(self) -> T:
        return self[0]

    def err(self) -> None:
        return self[1]


class Err(tuple, Generic[E, T]):
    is_ok: Final[TypeGuard["Ok"]] = False
    is_err: Final[TypeGuard["Err"]] = True

    def __new__(cls, e: E, t: Optional[T] = None):
        return tuple.__new__(cls, (t, e))

    def ok(self) -> Optional[T]:
        return self[0]

    def err(self) -> E:
        return self[1]


Result: TypeAlias = Ok[T] | Err[E, T]



class Debug:
    def __init__(self, logger) -> None:
        self._logger = logger
        self._indent = 0
        self._putsbuff = ""

    def indent(self, n=1) -> None:
        self._indent += n

    def undent(self, n=1) -> None:
        self._indent -= n

    def print(self, message, indent=0) -> None:
        space = ""
        for _ in range(self._indent + indent):
            space += "  "

        message = self._putsbuff + message
        self._putsbuff = ""

        lines = message.split("\n")
        for line in lines:
            self._logger.debug(space + line)

    def puts(self, message) -> None:
        self._putsbuff += message

        lines = self._putsbuff.split("\n")
        if len(lines) > 1:
            self._putsbuff = ""

            for line in lines[:-1]:
                self.print(line)

            self._putsbuff = lines[-1]



#
# path/to/file.md:ID.Id[name].id(tag,S,t:gd1002)
#


class Symbol:
    def __init__(self, symbol) -> None:
        pass

    @classmethod
    def isValid(cls, symbol):
        return True

    @classmethod
    def getFilePath(cls, symbol):
        parts = symbol.split(":")

        if len(parts) == 1:
            return None

        elif len(parts) == 2:
            return parts[0]

        else:
            # should raise.
            return None

    @classmethod
    def getId(cls, symbol):
        parts = symbol.split(":")

        if len(parts) <= 2:
            part = parts[-1]

        else:
            # should raise.
            return None

        parts = part.split("(")
        part = parts[0]

        if len(parts) == 2:
            if not parts[1].endswith(")"):
                # should raise.
                return None

        elif len(parts) > 2:
            # should raise.
            return None

        return part

    @classmethod
    def getTags(cls, symbol):
        tags = []
        parts = symbol.split("(")

        if len(parts) == 2:
            part = parts[1]
            if not part.endswith(")"):
                # should raise.
                return None

            tags = [x.strip() for x in part[:-1].split(",")]

        elif len(parts) > 2:
            # should raise.
            return None

        return tags

    @classmethod
    def isNameIncluded(cls, symbol):
        return True

    @classmethod
    def getParts(cls, symbol):
        parts = {"filePath": None, "symbol": [], "tag": []}

        return parts
